/*
 * uartlib.c
 *
 *  Created on: 08.07.2011
 *      Author: jmm
 */

#include <avr/interrupt.h>
#include "uartlib.h"

volatile uint8_t TXBuffer[UART_BUFFER_SIZE];
volatile uint8_t tx_start = 0;
volatile uint8_t tx_end = 1;

volatile uint8_t RXBuffer[UART_BUFFER_SIZE];
volatile uint8_t rx_start = 0;
volatile uint8_t rx_end = 1;

inline uint8_t bufferEmpty(uint8_t start, uint8_t end) {
	return (((start + 1) % UART_BUFFER_SIZE) == end);
}

inline uint8_t bufferFull(uint8_t start, uint8_t end) {
	return (((end + 1) % UART_BUFFER_SIZE) == start);
}

inline void enQueue(uint8_t value, volatile uint8_t buffer[], uint8_t end) {
	buffer[end] = value;
	end = ((end + 1) % UART_BUFFER_SIZE);
}

inline uint8_t deQueue(volatile uint8_t buffer[], uint8_t start) {
	start = ((start + 1) % UART_BUFFER_SIZE);
	return buffer[start];
}

ISR(USART_RXC_vect) {
	// Store UDR value
	if(!bufferFull(rx_start, rx_end))
		enQueue(UDR, RXBuffer, rx_end);
}

ISR(USART_UDRE_vect) {
	// Send a value
	if(!bufferEmpty(tx_start, tx_end))
		UDR = deQueue(TXBuffer, tx_start);
}

#if UART_FRAME_8N1
  void uart_init(uint32_t baud, uint8_t dataDirection) {
	  	uint16_t UBRR_VAL = ((F_CPU + baud * 8UL) / (baud * 16UL) - 1UL);
	  	UBRRH = (uint8_t) (UBRR_VAL >> 8);
	  	UBRRL = (uint8_t) UBRR_VAL;
	  	UCSRB = dataDirection;
	  	UCSRB |= (1 << RXCIE) || (1 << UDRIE);
		while ( !( UCSRA & (1<<UDRE)) );
		UDR = 255;
  }
#else
  void uart_init(uint32_t baud, uint8_t dataDirection, uint8_t charSize, uint8_t parity, uint8_t stopBits) {
  	uint16_t UBRR_VAL = ((F_CPU + baud * 8UL) / (baud * 16UL) - 1UL);
  	UBRRH = (uint8_t) (UBRR_VAL >> 8);
  	UBRRL = (uint8_t) UBRR_VAL;
  	UCSRB = dataDirection | (((charSize & 4) == 4) << UCSZ2);
  	UCSRC = (1 << URSEL) | parity | (stopBits << USBS) | (((charSize & 2) == 2) << UCSZ1) | (((charSize & 1) == 1) << UCSZ0);
  	UCSRB |= (1 << RXCIE) || (1 << UDRIE);
  }
#endif

inline void uart_transmit(uint8_t data) {
	while (bufferFull(tx_start, tx_end));
	enQueue(data, TXBuffer, tx_end);
}

inline uint8_t uart_receive() {
	while (bufferEmpty(rx_start, rx_end));
	uint8_t rxn = deQueue(RXBuffer, rx_start);
	return rxn;
}

inline void uart_transmitS(char* data) {
	while(*data)
		uart_transmit(*data++);
}

#if UART_FUNC_16B
	void uart_transmit16(uint16_t data) {
		uart_transmit(data >> 8);
		uart_transmit(data);
	}

	uint16_t uart_receive16() {
		uint16_t data = uart_receive(); data <<= 8;
		data |= uart_receive();
		return data;
	}
#endif

#if UART_FUNC_32B
	void uart_transmit32(uint32_t data) {
		uart_transmit(data >> 24);
		uart_transmit(data >> 16);
		uart_transmit(data >> 8);
		uart_transmit(data);
	}

	uint32_t uart_receive32() {
		uint32_t data = uart_receive(); data <<= 8;
		data |= uart_receive(); data <<= 8;
		data |= uart_receive(); data <<= 8;
		data |= uart_receive();
		return data;
	}
#endif

#if UART_FUNC_64B
	void uart_transmit64(uint64_t data) {
		uart_transmit(data >> 56);
		uart_transmit(data >> 48);
		uart_transmit(data >> 40);
		uart_transmit(data >> 32);
		uart_transmit(data >> 24);
		uart_transmit(data >> 16);
		uart_transmit(data >> 8);
		uart_transmit(data);
	}

	uint64_t uart_receive64() {
		uint64_t data = uart_receive(); data <<= 8;
		data |= uart_receive(); data <<= 8;
		data |= uart_receive(); data <<= 8;
		data |= uart_receive(); data <<= 8;
		data |= uart_receive(); data <<= 8;
		data |= uart_receive(); data <<= 8;
		data |= uart_receive(); data <<= 8;
		data |= uart_receive();
		return data;
	}
#endif

#if UART_FUNC_P
	void uart_transmitP(uint8_t* data) {
		uart_transmit(*data);
	}

	#if UART_FUNC_16B
	void uart_transmit16P(uint16_t* data) {
		uint8_t* dataX = (uint8_t*) data;
		uart_transmit(*(dataX + 1));
		uart_transmit(*(dataX + 0));
	}
	#endif

	#if UART_FUNC_32B
	void uart_transmit32P(uint32_t* data) {
		uint8_t* dataX = (uint8_t*) data;
		uart_transmit(*(dataX + 3));
		uart_transmit(*(dataX + 2));
		uart_transmit(*(dataX + 1));
		uart_transmit(*(dataX + 0));
	}
	#endif

	#if UART_FUNC_64B
	void uart_transmit64P(uint64_t* data) {
		uint8_t* dataX = (uint8_t*) data;
		uart_transmit(*(dataX + 7));
		uart_transmit(*(dataX + 6));
		uart_transmit(*(dataX + 5));
		uart_transmit(*(dataX + 4));
		uart_transmit(*(dataX + 3));
		uart_transmit(*(dataX + 2));
		uart_transmit(*(dataX + 1));
		uart_transmit(*(dataX + 0));
	}
	#endif
#endif
